/*  Project:		PROCEDURAL SMOKE
	---------------------------------------------------------------------------------------
	Author:			Anderes Hagvall
	Created:		June 03, 2007
	Description:	Point Sprite Application with rough smoke simulation using random noise
	---------------------------------------------------------------------------------------
*/

#define STRICT
#define WIN32_LEAN_AND_MEAN

//////////////
// inlcudes //
//////////////
#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include <mmsystem.h>	//timing
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glaux.h>
//#include "resource.h"
#include "vector3f.h"
#include "glext.h"		//local vendor dependent file


using namespace std;

////////////////////////////////////
// OpenGL point sprite extensions //
////////////////////////////////////
PFNGLPOINTPARAMETERFARBPROC  glPointParameterfARB  = NULL;
PFNGLPOINTPARAMETERFVARBPROC glPointParameterfvARB = NULL;

/////////////
// globals //
/////////////
HWND   g_hWnd      = NULL;
HDC	   g_hDC       = NULL;
HGLRC  g_hRC       = NULL;
GLuint g_textureID = -1;
bool   keys[256];
bool   keypress = false;

const float PI = 3.141592654f;

// simulation parameters
float	g_power = -0.2f;				//x init
float	g_pitch = 0.2f;					//y init
float	g_yaw = 0.3f;					//z init
float	g_variance = 0.1f;				//distorsion
float	g_restYspeed = 0.3f;			//final y speed
float	g_respawnAge = 3.0f;			//time to live
float	g_alphaRetard = 0.1f;			//deceleration constant
int		g_numberOfParticles = 10000;		//how many particles to render
int		g_numberOfKParticles = 10;
bool	g_alphaFade = false;			//using alpha fade or not?
float	g_constAlpha = 0.2f;			//for g_alphaFade = false
float	g_maxParticleSize = 35.0f;		//biggest possible particle
float	g_exhaustSize = 0.1f;
float   g_respawnX = 5.0f;
float   g_camPos = -10.0f;
float   g_camRot = 0.0f;
float	g_fSpinX = 0.0f;
float	g_fSpinY = 0.0f;
int		g_nrOfScreenShots = 0;
bool    g_fadeup = false;

// smoke particle
struct Particle
{
    vector3f	pos;
    vector3f	vel;
	float		ageAffection;
	float		age;
	float		r, g, b, alpha;
	int			exhaust;
};

// array of smoke particles
const long MAX_PARTICLES = 98000;
Particle g_particles[MAX_PARTICLES];
Particle g_particles2[MAX_PARTICLES];

///////////////////////////
// function declarations //
///////////////////////////

int WINAPI WinMain(HINSTANCE hInstance,HINSTANCE hPrevInstance, 
				   LPSTR lpCmdLine, int nCmdShow);
LRESULT CALLBACK WindowProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
void init(void);
void render(void);
void shutDown(void);
void initPointSprites(void);
void updatePointSprites(void);
void renderPointSprites(int i);
float getRandomMinMax(float fMin, float fMax);

bool sg_glDumpWindow(const char *filename, int win_width, int win_height);
bool sg_glWritePPMFile(const char *filename, GLubyte *buffer, int win_width, int win_height);


vector3f getRandomVector(void);


//////////////////////////////
// function implementations //
//////////////////////////////

///////////////////
// main function //
///////////////////
int WINAPI WinMain(	HINSTANCE hInstance,
					HINSTANCE hPrevInstance,
					LPSTR     lpCmdLine,
					int       nCmdShow )
{
	WNDCLASSEX winClass; 
	MSG        uMsg;

    memset(&uMsg,0,sizeof(uMsg));

	winClass.lpszClassName = "MY_WINDOWS_CLASS";
	winClass.cbSize        = sizeof(WNDCLASSEX);
	winClass.style         = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	winClass.lpfnWndProc   = WindowProc;
	winClass.hInstance     = hInstance;
    winClass.hIcon	       = LoadIcon(NULL,IDI_APPLICATION);//LoadIcon(hInstance, (LPCTSTR)IDI_OPENGL_ICON);
    winClass.hIconSm	   = LoadIcon(NULL,IDI_APPLICATION);//LoadIcon(hInstance, (LPCTSTR)IDI_OPENGL_ICON);
	winClass.hCursor       = LoadCursor(NULL, IDC_ARROW);
	winClass.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
	winClass.lpszMenuName  = NULL;
	winClass.cbClsExtra    = 0;
	winClass.cbWndExtra    = 0;
	
	if( !RegisterClassEx(&winClass) )
		return E_FAIL;

	g_hWnd = CreateWindowEx( NULL, "MY_WINDOWS_CLASS", 
                             "Procedural Smoke Generation",
						     WS_OVERLAPPEDWINDOW | WS_VISIBLE,
					         0, 0, 1400, 1000, NULL, NULL, hInstance, NULL );

	if( g_hWnd == NULL )
		return E_FAIL;

    ShowWindow( g_hWnd, nCmdShow );
    UpdateWindow( g_hWnd );

	init();
    initPointSprites();

	while( uMsg.message != WM_QUIT )
	{
		if( PeekMessage( &uMsg, NULL, 0, 0, PM_REMOVE ) )
		{
			TranslateMessage( &uMsg );
			DispatchMessage( &uMsg );
		}
        else // in this else statement, all action after initialization takes place
			if (keypress)
			{
				if (keys['Q']) {
					keys['Q']=false;
					g_power-=(g_power < -15.0f ? 0.0f : 0.5f );	
				}	//add exhaust
				else if (keys['A'])	{
					keys['A']=false;
					g_power+=(g_power > 0.5f ? 0.0f : 0.5f );
				}	//sub exhaust
				else if (keys['W']) {
					keys['W']=false;
					g_pitch+=(g_pitch > 4.0f ? 0.0f : 0.2f);
				}	//add y init
				else if (keys['S']) {
					keys['S']=false;
					g_pitch-=(g_pitch < -4.0f ? 0.0f : 0.2f);
				}	//sub y init
				else if (keys['E']) {
					keys['E']=false;
					g_yaw+=(g_yaw > 4.0f ? 0.0f : 0.2f);											
				}	//add yaw
				else if (keys['D']) {
					keys['D']=false;
					g_yaw-=(g_yaw < -4.0f ? 0.0f : 0.2f);											
				}	//sub yaw
				else if (keys['R']) {
					keys['R']=false;
					g_variance+=(g_variance > 2.0f ? 0.0f : 0.1f);	
				}	//add var
				else if (keys['F']) {
					keys['F']=false;
					g_variance-=(g_variance < 0.2f ? 0.0f : 0.1f);
				}	//sub var
				else if (keys['T']) {
					keys['T']=false;
					g_restYspeed+=(g_restYspeed > 3.0f ? 0.0f : 0.3f);
				}	//add y rest
				else if (keys['G']) {
					keys['G']=false;
					g_restYspeed-=(g_restYspeed < -3.0f ? 0.0f : 0.3f);
				}	//sub y rest
				else if (keys['Y']) {
					keys['Y']=false;
					g_respawnAge+=(g_respawnAge > 10.0f ? 0.0f : 0.5f);	
				}	//add particle life
				else if (keys['H']) {
					keys['H']=false;
					g_respawnAge-=(g_respawnAge < 0.5f ? 0.0f : 0.5f);
				}	//sub particle life
				else if (keys['U']) {
					keys['U']=false;
					g_alphaRetard+=(g_alphaRetard > 3.0f ? 0.0f : 0.1f);
				}	//add alpha retard
				else if (keys['J']) {
					keys['J']=false;
					g_alphaRetard-=(g_alphaRetard < 0.0f ? 0.0f : 0.1f);
				}	//sub alpha retard
				else if (keys['I']) {
					keys['I']=false;
					g_numberOfParticles+=(g_numberOfParticles >= MAX_PARTICLES ? 0 : 2000);
					g_numberOfKParticles = g_numberOfParticles/1000;
				}	//add particles
				else if (keys['K']) {
					keys['K']=false;
					g_numberOfParticles-=(g_numberOfParticles < 2000 ? 0 : 2000);
					g_numberOfKParticles = g_numberOfParticles/1000;
				}	//sub particles
				else if (keys['O']) {
					keys['O']=false;
					g_constAlpha+=(g_constAlpha > 1.0f ? 0.0f : 0.02f);	
				}	//add const alpha
				else if (keys['L']) {
					keys['L']=false;
					g_constAlpha-=(g_constAlpha < 0.0f ? 0.0f : 0.02f);
				}	//sub const alpha
				else if (keys['Z']) {
					keys['Z']=false;
					g_maxParticleSize+=(g_maxParticleSize > 1000.0f ? 0.0f : 4.0f);
				}	//bigger particles
				else if (keys['X']) {
					keys['X']=false;
					g_maxParticleSize-=(g_maxParticleSize < 4.0f ? 0.0f : 4.0f);
				}	//smaller particles
				else if (keys['C']) {
					keys['C']=false;
					g_alphaFade = !g_alphaFade;
				}	//change particle falloff
				else if (keys['V']) {
					keys['V']=false;
					g_exhaustSize += (g_exhaustSize > 1.0f ? 0.0f : 0.1f);
				}	//change exhause radius
				else if (keys['B']) {
					keys['B']=false;
					g_exhaustSize -= (g_exhaustSize < 0.0f ? 0.0f : 0.1f);
				}	//change exhause radius
				else if (keys[VK_F1]) {
					keys[VK_F1]=false;
					g_respawnX += (g_respawnX > 6.0f ? 0.0f : 0.2f);
				}	//change respawn pos
				else if (keys[VK_F2]) {
					keys[VK_F2]=false;
					g_respawnX -= (g_respawnX < 0.0f ? 0.0f : 0.2f);
				}	//change respawnpos
				else if (keys[VK_F5]) {
					keys[VK_F5]=false;
					g_camPos -= (g_camPos < -50.0f ? 0.0f : 0.5f);
				}	//change cam pos
				else if (keys[VK_F6]) {
					keys[VK_F6]=false;
					g_camPos += (g_camPos > -1.0f ? 0.0f : 0.5f);
				}	//change cam pos
				else if (keys[VK_F7]) {
					keys[VK_F7]=false;
					g_camRot += 5.0f;
				}	//change cam rot
				else if (keys[VK_F8]) {
					keys[VK_F8]=false;
					g_camRot -= 5.0f;
				}	//change cam rot
				else if (keys[VK_TAB]) {
					keys[VK_TAB]=false;
					const char* file = "screenshot0.ppm";
					g_nrOfScreenShots++;
					bool shot = sg_glDumpWindow(file, 800, 800);
				}	//take screen shot
				else if (keys['M']) {
					keys['M']=false;
					g_fadeup = !g_fadeup;
				}
				keypress = false;
			}
			render();
	}

	shutDown();

    UnregisterClass( "MY_WINDOWS_CLASS", hInstance );

	return uMsg.wParam;
}

//////////////////////
// Screenshot Stuff //
//////////////////////
bool sg_glWritePPMFile(const char *filename, GLubyte *buffer, int win_width, int win_height)
{
	int i, j, k, q;
    unsigned char *ibuffer;
    FILE *fp;
    int pixelSize = 4;

    ibuffer = (unsigned char *) malloc(win_width*win_height*3);

    if ( (fp = fopen(filename, "wb")) == NULL ) {
		printf("Warning: cannot open %s\n", filename);
		return false;
    }
 
    fprintf(fp, "P6\n# CREATOR: glReadPixel()\n%d %d\n%d\n",win_width, win_height, UCHAR_MAX);
    q = 0;
    for (i = 0; i < win_height; i++)
		for (j = 0; j < win_width; j++)
			for (k = 0; k < 3; k++)
				ibuffer[q++] = (unsigned char)*(buffer + (pixelSize*((win_height-1-i)*win_width+j)+k));
    fwrite(ibuffer, sizeof(unsigned char), 3*win_width*win_height, fp);
    fclose(fp);
    free(ibuffer);

    printf("wrote file (%d x %d pixels, %d bytes)\n",win_width, win_height, 3*win_width*win_height);
    return true;
}
 
 
// dump the screen buffer to a ppm file
bool sg_glDumpWindow(const char *filename, int win_width, int win_height) {
	GLubyte *buffer;
    bool result;
 
    buffer = (GLubyte *) malloc(win_width*win_height*4);
 
    // read window contents from color buffer with glReadPixels
    glFinish();
    glReadPixels(0, 0, win_width, win_height, GL_RGBA, GL_UNSIGNED_BYTE, buffer);
    result = sg_glWritePPMFile( filename, buffer, win_width, win_height );
    free(buffer);
 
    return result;
}

/////////////////////////////
// windows message handler //
/////////////////////////////
LRESULT CALLBACK WindowProc( HWND   hWnd, 
							 UINT   msg, 
							 WPARAM wParam, 
							 LPARAM lParam )
{
    static POINT ptLastMousePosit;
	static POINT ptCurrentMousePosit;
	static bool bMousing;

    switch( msg )
	{
        case WM_KEYDOWN: // a key gets pressed, handled in main loop though
		{
			keys[wParam] = true;
			keypress = true;
		}
		break;
		case WM_KEYUP:
		{
			keys[wParam] = false;
			keypress = false;
		}
        break;

        case WM_LBUTTONDOWN:
		{
			ptLastMousePosit.x = ptCurrentMousePosit.x = LOWORD (lParam);
            ptLastMousePosit.y = ptCurrentMousePosit.y = HIWORD (lParam);
			bMousing = true;
		}
		break;

		case WM_LBUTTONUP:
		{
			bMousing = false;
		}
		break;

		case WM_MOUSEMOVE:
		{
			ptCurrentMousePosit.x = LOWORD (lParam);
			ptCurrentMousePosit.y = HIWORD (lParam);

			if( bMousing )
			{
				g_fSpinX -= (ptCurrentMousePosit.x - ptLastMousePosit.x);
				g_fSpinY -= (ptCurrentMousePosit.y - ptLastMousePosit.y);
			}
			
			ptLastMousePosit.x = ptCurrentMousePosit.x;
            ptLastMousePosit.y = ptCurrentMousePosit.y;
		}
		break;

        case WM_SIZE:
		{
			int nWidth  = LOWORD(lParam); 
			int nHeight = HIWORD(lParam);
			glViewport(0, 0, nWidth, nHeight);

			glMatrixMode( GL_PROJECTION );
			glLoadIdentity();
			gluPerspective( 45.0, (GLdouble)nWidth / (GLdouble)nHeight, 0.1, 100.0);
		}
		break;

		case WM_CLOSE:
		{
			PostQuitMessage(0);	
		}

        case WM_DESTROY:
		{
            PostQuitMessage(0);
		}
        break;
		
		default:
		{
			return DefWindowProc( hWnd, msg, wParam, lParam );
		}
		break;
	}

	return 0;
}

////////////////////
// initialization //
////////////////////
void init( void )
{
	GLuint PixelFormat;

	PIXELFORMATDESCRIPTOR pfd;
	memset(&pfd, 0, sizeof(PIXELFORMATDESCRIPTOR));

    pfd.nSize      = sizeof(PIXELFORMATDESCRIPTOR);
    pfd.nVersion   = 1;
    pfd.dwFlags    = PFD_DRAW_TO_WINDOW |PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
    pfd.iPixelType = PFD_TYPE_RGBA;
    pfd.cColorBits = 16;
    pfd.cDepthBits = 16;
	
	g_hDC = GetDC( g_hWnd );
	PixelFormat = ChoosePixelFormat( g_hDC, &pfd );
	SetPixelFormat( g_hDC, PixelFormat, &pfd);
	g_hRC = wglCreateContext( g_hDC );
	wglMakeCurrent( g_hDC, g_hRC );

	glClearColor( 0.0f, 0.0f, 0.0f, 1.0f );
	glEnable(GL_TEXTURE_2D);

	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 45.0f, 640.0f / 480.0f, 0.1f, 100.0f);

    char *ext = (char*)glGetString( GL_EXTENSIONS );

    if( strstr( ext, "GL_ARB_point_parameters" ) == NULL )
    {
        MessageBox(NULL,"GL_ARB_point_parameters extension was not found",
            "ERROR",MB_OK|MB_ICONEXCLAMATION);
        return;
    }
    else // try to set point sprite extensions
    {
        glPointParameterfARB  = (PFNGLPOINTPARAMETERFARBPROC)wglGetProcAddress("glPointParameterfARB");
        glPointParameterfvARB = (PFNGLPOINTPARAMETERFVARBPROC)wglGetProcAddress("glPointParameterfvARB");

        if( !glPointParameterfARB || !glPointParameterfvARB )
        {
            MessageBox(NULL,"One or more GL_ARB_point_parameters functions were not found",
                "ERROR",MB_OK|MB_ICONEXCLAMATION);
            return;
        }
    }
}

///////////////
// shut down //
///////////////
void shutDown( void )	
{
    glDeleteTextures( 1, &g_textureID );

	if( g_hRC != NULL )
	{
		wglMakeCurrent( NULL, NULL );
		wglDeleteContext( g_hRC );
		g_hRC = NULL;
	}

	if( g_hDC != NULL )
	{
		ReleaseDC( g_hWnd, g_hDC );
		g_hDC = NULL;
	}
}

//////////////////
// random noise //
//////////////////
float getRandomMinMax( float fMin, float fMax )
{
    float fRandNum = (float)rand () / RAND_MAX;
    return fMin + (fMax - fMin) * fRandNum;
}

/////////////////////
// same for vector //
/////////////////////
vector3f getRandomVector( void )
{
	vector3f vVector;

    // Pick a random Z between -1.0f and 1.0f.
    vVector.z = getRandomMinMax( -1.0f, 1.0f );
    
    // Get radius of this circle
    float radius = (float)sqrt(1 - vVector.z * vVector.z);
    
    // Pick a random point on a circle.
    float t = getRandomMinMax( -PI, PI );

    // Compute matching X and Y for our Z.
    vVector.x = (float)cosf(t) * radius;
    vVector.y = (float)sinf(t) * radius;

	return vVector;
}

/////////////////////////////////
// Point Sprite Initialization //
/////////////////////////////////
void initPointSprites( void )
{
	// beer me a texture
	AUX_RGBImageRec *pTextureImage = auxDIBImageLoad( "particle3.bmp" );

    if( pTextureImage != NULL )
	{
        glGenTextures( 1, &g_textureID );

		glBindTexture( GL_TEXTURE_2D, g_textureID );

		glTexParameteri( GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_LINEAR );
		glTexParameteri( GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_LINEAR );

		glTexImage2D( GL_TEXTURE_2D, 0, 3, pTextureImage->sizeX, pTextureImage->sizeY, 0,
				     GL_RGB, GL_UNSIGNED_BYTE, pTextureImage->data );
	}

	if( pTextureImage )
	{
		if( pTextureImage->data )
			free( pTextureImage->data );

		free( pTextureImage );
	}

    // Query for the max point size supported by the hardware
    float maxSize = 0.0f;
    glGetFloatv( GL_POINT_SIZE_MAX_ARB, &maxSize );
	if (maxSize > g_maxParticleSize)
		maxSize = g_maxParticleSize;
    glPointSize( maxSize );

	// init particles
	for( int i = 0; i < MAX_PARTICLES; ++i )
    {
		/* algorithm:
		   - sample starting position in yz
		   - generate starting velocity
		   - generate ageaffection */

		float angle = float(rand()%360);
		float rad = getRandomMinMax(0.5f,1.0f)*g_exhaustSize;
		float x = g_respawnX;
		float y = rad+rad*sinf(angle);
		float z = rad*cosf(angle); // sample every other to right and left exhaust

		int exhaust = 0;  //= i%7;
		switch(exhaust)
		{
		case 0:
			g_particles[i].pos = vector3f(x,y,z);
			g_particles[i].vel = vector3f(g_power+getRandomMinMax(-g_variance, g_variance), 
									  getRandomMinMax(-g_pitch, g_pitch), 
									  getRandomMinMax(-g_yaw, g_yaw));
			g_particles[i].exhaust = 0;
			break;
		case 1:
			g_particles[i].pos = vector3f(rad,2,z);
			g_particles[i].vel = vector3f(-0.4f*g_power+getRandomMinMax(-g_pitch, g_pitch),
									  0.5f*g_power+getRandomMinMax(-g_variance, g_variance), 
									  getRandomMinMax(-g_yaw, g_yaw));
			g_particles[i].exhaust = 0;
			break;
		case 2:
			g_particles[i].pos = vector3f(rad,-2,z);
			g_particles[i].vel = vector3f(-0.4f*g_power+getRandomMinMax(-g_pitch, g_pitch),
									  -0.5f*g_power+getRandomMinMax(-g_variance, g_variance), 
									  getRandomMinMax(-g_yaw, g_yaw));
			g_particles[i].exhaust = 0;
			break;
		case 3:
			g_particles[i].pos = vector3f(-3.0f,y,z);
			g_particles[i].vel = vector3f(-0.5f*g_power+getRandomMinMax(-g_variance, g_variance),
									  getRandomMinMax(-g_pitch, g_pitch), 
									  getRandomMinMax(-g_yaw, g_yaw));
			g_particles[i].exhaust = 0;
			break;
		case 4:
			g_particles[i].pos = vector3f(x,y,z);
			g_particles[i].vel = vector3f(g_power+getRandomMinMax(-g_variance, g_variance), 
									  getRandomMinMax(-g_pitch, g_pitch), 
									  getRandomMinMax(-g_yaw, g_yaw));
			g_particles[i].exhaust = 0;
			break;
		case 5:
			g_particles[i].pos = vector3f(x,y,z);
			g_particles[i].vel = vector3f(g_power+getRandomMinMax(-g_variance, g_variance), 
									  getRandomMinMax(-g_pitch, g_pitch), 
									  getRandomMinMax(-g_yaw, g_yaw));
			g_particles[i].exhaust = 0;
			break;
		case 6:
			g_particles[i].pos = vector3f(x,y,z);
			g_particles[i].vel = vector3f(g_power+getRandomMinMax(-g_variance, g_variance), 
									  getRandomMinMax(-g_pitch, g_pitch), 
									  getRandomMinMax(-g_yaw, g_yaw));
			g_particles[i].exhaust = 0;
			break;
		}
        
        g_particles[i].r = 0.5f; //getRandomMinMax( 0.0f, 1.0f );
        g_particles[i].g = 0.5f; //getRandomMinMax( 0.0f, 1.0f );
        g_particles[i].b = 0.5f; //getRandomMinMax( 0.0f, 1.0f );

		g_particles[i].ageAffection = getRandomMinMax(0.5f, 2.0f);
		g_particles[i].age = 0.0f;
	}
}
//////////////////////////
// Update Point Sprites //
//////////////////////////
void updatePointSprites( void )
{
	// overall application time
	static double dStartAppTime = timeGetTime();
	float fElpasedAppTime = (float)((timeGetTime() - dStartAppTime) * 0.001);

	// get dt
	static double dLastFrameTime = timeGetTime();
	double dCurrenFrameTime = timeGetTime();
	float timeDelta = (float)((dCurrenFrameTime - dLastFrameTime) * 0.001);
	timeDelta = min(timeDelta, float(1.0f/30.0f));
	dLastFrameTime = dCurrenFrameTime;

	// move the particles
	for( int i = 0; i < g_numberOfParticles; ++i )
	{
		// update position
		g_particles[i].pos += g_particles[i].vel * (float)timeDelta;
		// update velocity according to noise and air
		g_particles[i].vel.x += getRandomMinMax(-0.1*g_variance, 0.1*g_variance);
		g_particles[i].vel.z += getRandomMinMax(-0.2f*g_variance, 0.2f*g_variance);
		g_particles[i].vel.y += getRandomMinMax(-0.2f*g_variance, 0.2f*g_variance);
		if (g_particles[i].exhaust == 3) g_particles[i].vel.y *= 0.2f;
		g_particles[i].vel.y -= (g_particles[i].vel.y - g_restYspeed)*timeDelta;
		g_particles[i].vel.x -= g_alphaRetard*timeDelta*g_particles[i].vel.x;
		//if (g_particles[i].vel.y < g_restYspeed)
		//	g_particles[i].vel.y += timeDelta;
		//else g_particles[i].vel.y -= timeDelta;

		g_particles[i].age += g_particles[i].ageAffection * timeDelta;
		
		// if old particle, respawn it
		if (g_particles[i].age > g_respawnAge)
		{
			float angle = float(rand()%360);
			float rad = getRandomMinMax(-1.0f,1.0f);
			if (rad > 0) rad = sqrt(sqrt(rad))*g_exhaustSize;
			else rad = -sqrt(sqrt(-rad))*g_exhaustSize;
			float x = g_respawnX;
			float y = rad+rad*sinf(angle);
			float z = rad*cosf(angle); // sample every other to right and left exhaust
			
			switch(g_particles[i].exhaust)
			{
			case 0:
				g_particles[i].pos = vector3f(x,y,z);
				g_particles[i].vel = vector3f(g_power+getRandomMinMax(-g_variance, g_variance), 
										  getRandomMinMax(-g_pitch, g_pitch), 
										  getRandomMinMax(-g_yaw, g_yaw));
				break;
			case 1:
				g_particles[i].pos = vector3f(rad-1,3,z);
				g_particles[i].vel = vector3f(getRandomMinMax(-g_pitch, g_pitch),
										  0.4f*g_power+getRandomMinMax(-g_variance, g_variance), 
										  getRandomMinMax(-g_yaw, g_yaw));
				break;
			case 2:
				g_particles[i].pos = vector3f(rad-1,-3,z);
				g_particles[i].vel = vector3f(getRandomMinMax(-g_pitch, g_pitch),
										  -0.4f*g_power+getRandomMinMax(-g_variance, g_variance), 
										  getRandomMinMax(-g_yaw, g_yaw));
				break;
			case 3:
				g_particles[i].pos = vector3f(-3.0f,y,z);
				g_particles[i].vel = vector3f(-0.5f*g_power+getRandomMinMax(-g_variance, g_variance),
										  0.1f*getRandomMinMax(-g_pitch, g_pitch), 
										  getRandomMinMax(-g_yaw, g_yaw));
				break;
			case 4:
				g_particles[i].pos = vector3f(-y-0.5f,-x,z);
				g_particles[i].vel = vector3f(getRandomMinMax(-g_pitch, g_pitch),
										  -g_power+getRandomMinMax(-g_variance, g_variance), 
										  getRandomMinMax(-g_yaw, g_yaw));
				break;
			case 5:
				g_particles[i].pos = vector3f(y-2,x,z);
				g_particles[i].vel = vector3f(getRandomMinMax(-g_pitch, g_pitch),
										  g_power+getRandomMinMax(-g_variance, g_variance), 
										  getRandomMinMax(-g_yaw, g_yaw));
				break;
			case 6:
				g_particles[i].pos = vector3f(-y-2,-x,z);
				g_particles[i].vel = vector3f(getRandomMinMax(-g_pitch, g_pitch),
										  -g_power+getRandomMinMax(-g_variance, g_variance), 
										  getRandomMinMax(-g_yaw, g_yaw));
				break;
			}
			
			g_particles[i].age = 0.0f;
		}
	}
}

//////////////////////////
// Render Point Sprites //
//////////////////////////
void renderPointSprites( int p )
{
    // Blending
	glEnable( GL_BLEND );
	glBlendFunc( GL_SRC_ALPHA, GL_ONE );

    // This is how will our point sprite's size will be modified by 
    // distance from the viewer
    float quadratic[] =  { 1.0f, 0.0f, 0.01f };
    glPointParameterfvARB( GL_POINT_DISTANCE_ATTENUATION_ARB, quadratic );

    // Query for the max point size supported by the hardware
    float maxSize = g_maxParticleSize;//0.0f;
    //glGetFloatv( GL_POINT_SIZE_MAX_ARB, &maxSize );

    // Clamp size 
	//if( maxSize > g_maxParticleSize )
    //    maxSize = g_maxParticleSize;

    glPointSize( maxSize );

    // The alpha of a point is calculated to allow the fading of points 
    // instead of shrinking them past a defined threshold size. The threshold 
    // is defined by GL_POINT_FADE_THRESHOLD_SIZE_ARB and is not clamped to 
    // the minimum and maximum point sizes.
    glPointParameterfARB( GL_POINT_FADE_THRESHOLD_SIZE_ARB, 60.0f );
    glPointParameterfARB( GL_POINT_SIZE_MIN_ARB, 1.0f );
    glPointParameterfARB( GL_POINT_SIZE_MAX_ARB, maxSize );

    // Specify point sprite texture coordinate replacement mode for each 
    // texture unit
    glTexEnvf( GL_POINT_SPRITE_ARB, GL_COORD_REPLACE_ARB, GL_TRUE );

	for (int j = 0; j < p; j++)
	{
		int k = 1000*j;
		// Render them particles
		glEnable( GL_POINT_SPRITE_ARB );
		
		glBegin( GL_POINTS );
		for( int i = k; i < k+1000; i++ )
		{
			glColor4f( g_particles[i].r, 
					   g_particles[i].g, 
					   g_particles[i].b, 
					   (g_fadeup ? g_constAlpha*g_particles[i].age/g_respawnAge : (g_alphaFade ? 1.0f-g_particles[i].age/g_respawnAge : g_constAlpha)));

			glVertex3f( g_particles[i].pos.x,
						g_particles[i].pos.y,
						g_particles[i].pos.z );
		}
		glEnd();
		glDisable( GL_POINT_SPRITE_ARB );
	}
}

//////////////////
// Scene Render //
//////////////////
void render( void )
{
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	
	glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
	glTranslatef( 0.0f, 0.0f, g_camPos );
    glRotatef( -g_fSpinY, 1.0f, 0.0f, 0.0f );
    glRotatef( -g_fSpinX, 0.0f, 1.0f, 0.0f );
	glRotatef( g_camRot, 0.0f, 1.0f, 0.0f );

    updatePointSprites();

	glBindTexture( GL_TEXTURE_2D, g_textureID );

	renderPointSprites(g_numberOfKParticles);

	SwapBuffers( g_hDC );
}